from .nda_functions import get_process_name
from .nda_functions import get_barcode
from .nda_functions import get_start_time
from .nda_functions import cycle
from .nda_functions import recipe
from .nda_functions import step
from .nda_functions import records



#!/bin/python3.4

# from audioop import mul
# import sys, getopt
import struct
# import binascii
# import time
import logging
import mmap
import re
import datetime
import pandas as pd
ILLEGAL_CHARACTERS_RE = re.compile(r'[\000-\010]|[\013-\014]|[\016-\037]')

def validate_timegap(df):

    
    """
    #:Check if there is a gap due to electricity cutting off and thus if there are errors there. 
    #:Can check if removing those records makes the dataframe pass the validation    
    """
    
    
    df['prev_step']=df['step_ID'].shift(periods=1)
    
    df['prev_tis']=df['time_in_step'].shift(periods=1)
    df['prev_tstamp']=df['timestamp'].shift(periods=1)
    df['prev_tis']=df['time_in_step']-df['prev_tis']
    df['prev_tstamp']=df['timestamp']-df['prev_tstamp']
    df['prev_tstamp']=df['prev_tstamp'].dt.total_seconds()
    

    #todo prev_tis and prev_tstamp now represent difference. of the time in step and timestamp 
    df.loc[( (abs(df['prev_tis']-df['prev_tstamp'])>5) & ((df['step_ID']==df['prev_step']) & (df['time_in_step']!=0)) ),'Validated']=True
    df.drop(columns=['prev_tis','prev_tstamp','prev_step'],inplace=True)
       
    









def main_validator(df,capacity_nom):
    """
    It checks if the dataframe is valid. 
    
    The function takes in a dataframe and a capacity_nom. 
    
    It checks if the record_ID is monotonically increasing and starts at 1. 
    
    It checks if the cycle starts at 1. 
    
    It checks if the step_ID is monotonically increasing and starts at 1. 
    
    It checks if the voltage is greater than 2. 
    
    It checks if the capacity is greater than 0. 
    
    It checks if the capacity is less than 1500 times the capacity_nom. 
    
    It checks if the current is less than 1600 times the capacity_nom. 
    
    It checks if the process name, start time, and barcode are valid. 
    
    It checks if the barcode is 12 characters long.
    
    Args:
      df: the dataframe to be validated
      capacity_nom: The nominal capacity of the battery.
    
    Returns:
      True if the dataframe is valid, and False if it is not.
    """
  
    validate_timegap(df)
    if(False in df.Validated.values):
        print('Df validation failed. Kindly check.')
        logging.info("This df has failed the basic validation after ignoring the timegaps' records")
        return False


    #!Could return numbers instead of True False to indicate what error we have. And then check for electricity timegap with that error
    df.loc[(df['step_name']=='Rest') & (df['current_mA']==0) & (df['time_in_step']!=0),'Validated']=False 
    if(df['record_ID'].min()!=1) or (not df['record_ID'].is_monotonic_increasing): 
        logging.info("Record_ID error. Min record_ID is: "+str(df['record_ID'].min()))
        return False
    if(df['cycle'].min()!=1): 
        logging.info("Cycle error. Min cycle is: "+str(df['cycle'].min()))
        return False
    if(df['step_ID'].min()!=1 or (not df['step_ID'].is_monotonic_increasing)): 
        logging.info("step_ID error")
        return False
    if((df['voltage_V'].min()<2)and (df[df['step_name']=='SIM'].shape[0]==0)): 
        logging.info("Voltage error. Min voltage is: "+str(df['voltage_V'].min()))
        return False
    if((df['energy_mWh'].min()<0) and (df[df['step_name']=='SIM'].shape[0]==0)):
        logging.info("Negative Energy error. Min energy is: "+str(df['energy_mWh'].min()))
        return False
    
    if((df['capacity_mAh'].min()<0) and (df[df['step_name']=='SIM'].shape[0]==0)):
        logging.info("Negative Capacity error. Min capacity is: "+str(df['capacity_mAh'].min()))
        return False
    
    if(df['capacity_mAh'].max()>(1500*capacity_nom)): 
        logging.info("Max Capacity error. Max capacity is: "+str(df['capacity_mAh'].max()))
        return False
    if(df['current_mA'].max()>(1600*capacity_nom)):
        logging.info("Current error. Max current is: "+str(df['current_mA'].max()))
        return False
    if all(ele in df.keys() for ele in ['Process Name','Start Time','barcode']):    
        if ILLEGAL_CHARACTERS_RE.search(df['Process Name'].unique()[0]):
            logging.info("Process name error. The decoded Process name is: "+str(df['Process Name'].unique()[0]))
            return False
        if ILLEGAL_CHARACTERS_RE.search(df['Start Time'].unique()[0]):
            logging.info("Start Time error. The decoded Start Time is: "+str(df['Start Time'].unique()[0]))
            return False
        if ILLEGAL_CHARACTERS_RE.search(df['barcode'].unique()[0]):
            logging.info("Barcode error. The decoded Barcode is: "+str(df['barcode'].unique()[0]))
            return False
        if(len(df['barcode'].unique()[0])!=12):
            logging.info("Barcode length error. The decoded Barcode is: "+str(df['barcode'].unique()[0]))
            return False
   
    return True

def get_process_name(inpath):
    with open(inpath, 'rb') as f:
        data = f.read()
        t = -1
        x8=data.find(b'BTS Client')
        for i in range(x8-160,x8-100):
            if data[i:i+1] == b'\x00':
                t = i
                break
        process_name=data[x8-160:t]
    return process_name.decode()


def get_st_time(inpath):
    with open(inpath, 'rb') as f:
        data = f.read()
        pos=data.find(b'BTS Client')

        st_time2=data[pos-566:pos-547].decode()


        return(st_time2)
    
def get_barcode(inpath):
    with open(inpath, 'rb') as f:
        data = f.read()
        t = -1
        x8=data.find(b'BTS Client')
        for i in range(x8-260,x8):
            if data[i:i+1] == b'\x00':
                t = i
                break
        barcode=data[x8-260:t]
    return barcode.decode()

def get_remarks(inpath):
    with open(inpath, 'rb') as f:
        data = f.read()
        t = -1
        x8=data.find(b'BTS Client')
        for i in range(x8-376,x8):
            if data[i:i+1] == b'\x00':
                t = i
                break
        remarks=data[x8-376:t]
    return remarks.decode()

def ValidFile(inpath):
    with open(inpath, 'rb') as f:
        data = f.read()
        pos=data.find(b'BTS Client')
        if(pos==-1):
            return False
        else:
            return True

        

def _count_changes(series):
    """Enumerate the number of value changes in a series"""
    a = series.diff()
    a.iloc[0] = 1
    a.iloc[-1] = 0
    return((abs(a) > 0).cumsum())




def single_validator(List):

    
    
# #!IMPORTRANT TEMPORARY CODE FOR TESTING PUPOSES PLS REMOVE    
    # return True
# #!IMPORTRANT TEMPORARY CODE FOR TESTING PUPOSES PLS REMOVE
    
    if(List[0]<1 or List[1]<1 or List[2]<1 or List[4]<0 or List[5]<2 or List[7]<0 or List[8]<0): 

        return False #Index/Record_ID #cycle #step_ID #time_in_step #voltage #capacity #energy
    if ILLEGAL_CHARACTERS_RE.search(List[3]):
        return False

    return True



# Return a dict containing the relevant data.  all nice and pretty like.
def _bytes_to_list(byte_stream):

    List=[]

    state_dict = {
        1: 'CC_Chg',
        2: 'CC_Dchg',
        3: 'CV_Chg',
        4: 'Rest',
        5: 'Cycle',
        7: 'CCCV_Chg',
        10: 'CR_Dchg',
        13: 'Pause',
        16: 'Pulse',
        17: 'SIM',
        19: 'CV_Dchg',
        20: 'CCCV_Dchg'
    }
    range_list=[200000,-100000,-60000,-30000,-50000,-20000,-12000,-10000,-6000,-5000,-3000,-1000,-500,-100,0,10,100,200,1000,6000,12000,50000,60000]
    multiplier_dict = {
        -200000: 1e-2,
        -100000: 1e-2,
        -60000: 1e-2,
        -12000: 1e-2,
        -30000: 1e-2,
        -50000: 1e-2,
        -20000: 1e-2,
        -10000: 1e-2, 
        
        -6000: 1e-2,
        -5000: 1e-2,
        -3000: 1e-2,
        -1000: 1e-2,
        -500: 1e-3,
        -100: 1e-3,
        0: 0,
        10: 1e-3,
        100: 1e-2,
        200: 1e-2,
        1000: 1e-1,
        6000: 1e-1,
        12000: 1e-1,
        50000: 1e-1,
        60000: 1e-1,
    }


    [Index, Cycle] = struct.unpack('<IH', byte_stream[2:8])
    # [Step] = struct.unpack('<I', byte_stream[10:14])
    [Status, Jump, Time] = struct.unpack('<BBQ', byte_stream[12:22])    #state_dict[Status] is step_name
    [Step] = struct.unpack('<H', byte_stream[10:12])
    [Voltage, Current] = struct.unpack('<ii', byte_stream[22:30])
    [Charge_capacity, Discharge_capacity] = struct.unpack('<qq', byte_stream[38:54])
    [Charge_energy, Discharge_energy] = struct.unpack('<qq', byte_stream[54:70])
    [Y, M, D, h, m, s] = struct.unpack('<HBBBBB', byte_stream[70:77])
    [Range] = struct.unpack('<i', byte_stream[78:82])

    try:
        Date = datetime.datetime(Y, M, D, h, m, s)

    except ValueError:
        [Timestamp] = struct.unpack('<Q', byte_stream[70:78])
        Date = datetime.datetime.fromtimestamp(Timestamp)

    if(Range in range_list):
        multiplier = multiplier_dict[Range]
    else:
        raise ValueError("At ",Index," the ",Range," range caused an error")
    
    List = [
        Index,
        Cycle + 1,
        Step,
        state_dict[Status],
        Time/1000,
        Voltage/10000,
        Current*multiplier,
        abs(Charge_capacity-Discharge_capacity) *multiplier/3600,
        abs(Charge_energy-Discharge_energy)*multiplier/3600,
        Date,
    ]

   
    List.append(single_validator(List))
    return List
    
        
    
def _aux_bytes_to_list(bytes):
    """Helper function for intepreting auxiliary records"""
    [Aux] = struct.unpack('<B', bytes[1:2])
    [Index, Cycle] = struct.unpack('<IB', bytes[2:7])
    [T] = struct.unpack('<h', bytes[34:36])

    list = [
        Index,
        Aux,
        T/10
    ]

    return(list)






def _valid_record(bytes):
    """Helper function to identify a valid record"""
    # Check for a non-zero Status
    [Status] = struct.unpack('<B', bytes[12:13])
    return(Status != 0)


#! Output for newest BTSDA version
def nda_in_df_out(file):
  
    
    rec_columns = [
    'record_ID', 'cycle' ,'step_ID','step_name', 'time_in_step', 'voltage_V',
    'current_mA', 'capacity_mAh','energy_mWh','timestamp','Validated']
    aux_columns = ['Index', 'Aux', 'T']

    record_len = 86
    main_data = False
    with open(file, "rb") as f:
        mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        mm_size = mm.size()

        if mm.read(6) != b'NEWARE':
            raise ValueError(f"{file} does not appear to be a Neware file.")

        if(mm.find(b'BTS Client')==-1):
            raise ValueError(f"{file} does not appear to be a correctly downloaded Neware file.")
    
        identifier = b'\x00\x00\x00\x00\x55\x00'
        header = mm.find(identifier)
        if header == -1:
            raise EOFError(f"File {file} does not contain any valid records.")
        while (((mm[header + 4 + record_len] != 85) | (not _valid_record(mm[header+4:header+4+record_len]))) if header + 4 + record_len < mm_size
               else False):
            header = mm.find(identifier, header + 4)
        mm.seek(header + 4)

        output = []
        aux = []
        while mm.tell() < mm_size:
            bytes = mm.read(record_len)
            if len(bytes) == record_len:

                # Check for a data record
                
                if (bytes[0:2] == b'\x55\x00'
                        and bytes[82:87] == b'\x00\x00\x00\x00'):
                                       
                    output.append( _bytes_to_list(bytes))

                # Check for an auxiliary record
                elif (bytes[0:1] == b'\x65'
                      and bytes[82:87] == b'\x00\x00\x00\x00'):
                    aux.append(_aux_bytes_to_list(bytes))

    df = pd.DataFrame(output, columns=rec_columns)
    
    df.dropna(inplace=True)
    df.drop_duplicates(inplace=True)
    
    if(get_barcode(file).startswith('0AD')):
        df.drop(df.index[df['step_ID']<7],axis=0,inplace=True)
    df.record_ID = _count_changes(df.record_ID)
    df.step_ID = _count_changes(df.step_ID)

    if not df.record_ID.is_monotonic_increasing:
        df.sort_values('record_ID', inplace=True)

    df.reset_index(drop=True, inplace=True)
    
    #todo sets timegap records Validation to true
    validate_timegap(df)
    
    # Join temperature data
    aux_df = pd.DataFrame(aux, columns=aux_columns)


   

    aux_df.drop_duplicates(inplace=True)
    if not aux_df.empty:
        for Aux in aux_df.Aux.unique():
            df = df.join(aux_df.loc[aux_df.Aux == Aux, 'T']) # type: ignore
            df.rename(columns={'T': f"T{Aux}"}, inplace=True)
   
    
    #!DCIR Calculation
    df['prev_cur']=df['current_mA'].shift(periods=1)
    df['prev_vol']=df['voltage_V'].shift(periods=1)
    df['DCIR']=-1
    df.loc[((df['prev_cur']==0)&(df['current_mA']!=0)),'DCIR']=abs((df['voltage_V']-df['prev_vol'])/(df['current_mA']-df['prev_cur']))*1000
    df.drop(columns=['prev_cur','prev_vol'],inplace=True)

    dtype_dict = {
        'record_ID': 'uint32',
        'cycle': 'uint32',
        'step_ID': 'uint32',
        'step_name': 'str',
        'time_in_step': 'uint32',
        'voltage_V': 'float32',
        'current_mA': 'float32',
        'capacity_mAh': 'float32',
        'energy_mWh': 'float32',
        'DCIR': 'float32',
        'Validated':'bool'
    }
    df = df.astype(dtype=dtype_dict)
    return df
   






# Libraries

from sys import displayhook
import pandas as pd
import numpy as np
from datetime import timedelta
import re
from . import nda_version_8_0


def _validator_cycle(df):
    '''
    To validate the data generated for CycleWise.

    Index(['Cycle_Index', 'Onset_Date', 'End_Date', 'Chg_Cap_(Ah)',
       'DChg_Cap_(Ah)', 'Chg_Energy_(Wh)', 'DChg_Energy_(Wh)',
       'Chg_Time(hh:mm:ss)', 'DChg_Time(hh:mm:ss)', 'Chg_Onset_Volt_(V)',
       'DChg_Oneset_Volt_(V)', 'End_of_Chg_Volt_(V)', 'End_of_DChg_Volt_(V)',
       'Chg_Oneset_Current_(A)', 'DChg_Oneset_Curent_(A)',
       'End_of_Chg_Current_(A)', 'End_of_DChg_Current_(A)', 'DCIR(mΩ)'],
       dtype='object')
    '''
    for id, i in df.iterrows():
        if (pd.to_datetime(i['End_Date']) > pd.to_datetime(i['Onset_Date'])):
            raise ValueError('Date error with cycle number '+str(i['Cycle_Index']))
        if (i["Cycle_Index"] < 1):
            raise ValueError('Cycle Index out of range; please check data.')
        if (i['Chg_Onset_Volt_(V)'] < 1.5 or i['DChg_Onset_Volt_(V)'] < 1.5):
            raise ValueError('Voltage value too low in cycle number '+str(i['Cycle_Index']))
        if (i['DCIR(mΩ)'] < 0):
            raise ValueError('DCIR is negative in cycle number '+str(i['Cycle_Index']))

def get_barcode(nda):
    '''
    returns barcode for said nda file
    '''
    if (nda.split('.')[-1] != 'nda'):
        raise ValueError("File passed in function is not an nda file")
    return nda_version_8_0.get_barcode(nda)


def get_start_time(nda):
    '''
    returns start time for the passed nda file
    '''
    if (nda.split('.')[-1] != 'nda'):
        raise ValueError("File passed in function is not an nda file")
    return nda_version_8_0.get_st_time(nda)


def get_process_name(nda):
    '''
    returns recipe name for said NDA file
    '''
    if (nda.split('.')[-1] != 'nda'):
        raise ValueError("File passed in function is not an nda file")
    return nda_version_8_0.get_process_name(nda)


def records(nda, rename=False):
    '''
    returns a Dataframe record-wise for the nda file
    Use the rename arguement if you want to rename the columns
    '''
    if (nda.split('.')[-1] != 'nda'):
        raise ValueError("File passed in function is not an nda file")
    df = nda_version_8_0.nda_in_df_out(nda)
    rec_columns = {
        'record_ID': 'DataPoint',
        'cycle': 'Cycle Index',
        'step_ID': 'Step Index',
        'step_name': 'Step Type',
        'time_in_step': 'Time',
        'voltage_V': 'Voltage(V)',
        'current_mA': 'Current(A)',
        'capacity_mAh': 'Capacity(Ah)',
        'energy_mWh': 'Energy(Wh)',
        'timestamp': 'Date',
        'Validated': 'Validated'
        }
    if (rename == True):
        df['current_mA'] = df['current_mA'].div(1000)
        df['capacity_mAh'] = df['capacity_mAh'].div(1000)
        df['energy_mWh'] = df['energy_mWh'].div(1000)
        df=df.rename(columns = rec_columns)
    return df



def cycle(df):  #! Function to group the data cycle-wise
    '''
    When passed an nda file or the records data,
    it returns Cycle-wise data identical to the
    cycle sheet in the excel file of the test.
    '''
    try:
        if type(df) != type(pd.DataFrame()):
            df = nda_version_8_0.nda_in_df_out(df)
    except Exception:
        raise ValueError('Arguement pushed into the function is neither a DataFrame nor a path to an nda file')

    temp_list=[]
    complete_list = []
    

    # df['prev_cur'] = df['current_mA'].shift(periods=1)
    # df['prev_vol']=df['voltage_V'].shift(periods=1)
    # df['DCIR']=-1
    # df.loc[((df['prev_cur']==0)&(df['current_mA']!=0)),'DCIR']=abs((df['voltage_V']-df['prev_vol'])/(df['current_mA']-df['prev_cur']))*1000
    # df.drop(columns=['prev_cur','prev_vol'],inplace=True)
    
    chg_temp = 'CCCV_Chg' #default values
    dchg_temp = "CC_Dchg"    
    step_col=list(df['step_name'].unique()[1:])
    for col in step_col:
        if re.search('chg',col,re.IGNORECASE):
            chg_temp=col
        if re.search('dchg',col,re.IGNORECASE):
            dchg_temp=col                           


    for i in range(min(df['cycle']),max(df['cycle']+1)):                         
        # df2=df.groupby('cycle').get_group(i)

        df2 = df[df['cycle']==i]
        cycle=i                                                                                                                                                                           

        starting_date=df2['timestamp'].head(1).values[0]
        end_date=df2['timestamp'].tail(1).values[0]
        
        df_chg=df2[(df2['step_name']==chg_temp)]
        df_dchg=df2[(df2['step_name']==dchg_temp)]

        charging_capacity=max(df_chg['capacity_mAh'],default=-1)
        discharging_capacity=max(df_dchg['capacity_mAh'],default=-1)

        charging_energy=max(df_chg['energy_mWh'],default=-1)
        discharging_energy=max(df_dchg['energy_mWh'],default=-1)

        chg_starting_volt=df_chg['voltage_V'].head(1).values[0]
        chg_ending_volt=df_chg['voltage_V'].tail(1).values[0]
        dchg_starting_volt=df_dchg['voltage_V'].head(1).values[0]
        dchg_ending_volt=df_dchg['voltage_V'].tail(1).values[0]

        chg_starting_current=df_chg['current_mA'].head(1).values[0]
        chg_ending_current=df_chg['current_mA'].tail(1).values[0]
        dchg_starting_current=df_dchg['current_mA'].head(1).values[0]
        dchg_ending_current=df_dchg['current_mA'].tail(1).values[0]

        charging_time=timedelta(seconds=list(df_chg['time_in_step'])[-1])
        discharging_time=timedelta(seconds=list(df_dchg['time_in_step'])[-1])

        dcir_cyl=df2.loc[df2['DCIR']>0,'DCIR']
    
        DCIR_avg=dcir_cyl.mean()     

        temp_list=[
                    cycle,
                    starting_date,
                    end_date,
                    charging_capacity,
                    discharging_capacity,
                    charging_energy,
                    discharging_energy,
                    charging_time,
                    discharging_time,
                    chg_starting_volt,
                    dchg_starting_volt,
                    chg_ending_volt,
                    dchg_ending_volt,
                    chg_starting_current,
                    dchg_starting_current,
                    chg_ending_current,
                    dchg_ending_current,
                    DCIR_avg
                    ]
        
        complete_list.append(temp_list)
    df3=pd.DataFrame(complete_list,columns=['Cycle Index','Onset Date','End Date','Chg. Cap.(Ah)','DChg. Cap.(Ah)',
                                                    'Chg. Energy(Wh)','DChg. Energy_(Wh)','Chg_Time(hh:mm:ss)','DChg_Time(hh:mm:ss)','Chg_Onset_Volt_(V)',
                                                    'DChg_Oneset_Volt_(V)','End_of_Chg_Volt_(V)','End_of_DChg_Volt_(V)','Chg_Oneset_Current_(A)','DChg_Oneset_Curent_(A)',
                                                    'End_of_Chg_Current_(A)','End_of_DChg_Current_(A)','DCIR(mΩ)'])
    return df3


def _validator_step(df) :   

    for id,i in df.iterrows():
        if(pd.to_datetime(i['End Date'])>pd.to_datetime(i['Onset Date'])):
            raise ValueError('Date error with cycle number '+str(i['Cycle_Index']))
        if(i["Cycle_Index"]<1):
            raise ValueError('Cycle Index out of range; please check data.')
        if(i['Chg_Onset_Volt_(V)']<1.5 or i['DChg_Onset_Volt_(V)']<1.5):
            raise ValueError('Voltage value too low in cycle number '+str(i['Cycle_Index']))
        if(i['DCIR(mΩ)']<0):
            raise ValueError('DCIR is negative in cycle number '+str(i['Cycle_Index']))
        # if


def step(df):  #! Function to group the data step-wise
    '''
    When passed an nda file or the records data,
    it returns Step-wise data identical to the
    cycle sheet in the excel file of the test.
    '''
    try:
        if type(df)!=type(pd.DataFrame()):
            df=nda_version_8_0.nda_in_df_out(df)
    except:
        raise ValueError('Arguement pushed into the function is neither a DataFrame nor a path to an nda file')
    # print(df)
    # df['current_mA']=df['current_mA'].div(1000)
    # df['capacity_mAh']=df['capacity_mAh'].div(1000)
    # df['energy_mWh']=df['energy_mWh'].div(1000)
    rec_columns = [
    'record_ID', 'cycle' ,'step_ID','step_name', 'time_in_step', 'voltage_V',
    'current_mA', 'capacity_mAh','energy_mWh','timestamp','Validated','DCIR']
    if(list(df.keys())!=rec_columns):
        raise ValueError ('DataFrame passed ')
        
    
    
    df['prev_cur']=df['current_mA'].shift(periods=1)
    df['prev_vol']=df['voltage_V'].shift(periods=1)
    df['prev_step']=df['step_ID'].shift(periods=1)


    df['DCIR']=-1
    df.loc[((df['prev_cur']==0)&(df['current_mA']!=0)&(df['prev_step']!=df['step_ID'])),'DCIR']=abs((df['voltage_V']-df['prev_vol'])/(df['current_mA']-df['prev_cur']))*1000
    df.drop(columns=['prev_cur','prev_vol','prev_step'],inplace=True)

    temp_list=[]
    complete_list=[]

    for i in range(min(df['step_ID']),max(df['step_ID'])+1):
        # print(i)
        # df2=df.groupby('step_ID').get_group(i)
        df2=df[df['step_ID']==i]
        DCIR=list(df2['DCIR'])[0]
        Starting_Volt=list(df2['voltage_V'])[0]
        End_Voltage=list(df2['voltage_V'])[-1]
        Starting_current=list(df2['current_mA'])[0]
        End_Current=list(df2['current_mA'])[-1]
        Capacity=list(df2['capacity_mAh'])[-1]
        Energy=list(df2['energy_mWh'])[-1]
        Starting_Date=list(df2['timestamp'])[0]
        End_Date=list(df2['timestamp'])[-1]
        Step_Time=timedelta(seconds=list(df2['time_in_step'])[-1])
        step_ID=i
        Maximum_voltage=max(list(df2['voltage_V']))
        Minimum_voltage=min(list(df2['voltage_V']))
        Step_Type=list(df2['step_name'])[0]
        Cycle_Index=list(df2['cycle'])[0]

        temp_list= [
            Cycle_Index,
            step_ID,
            Step_Type,
            str(Step_Time),
            Starting_Date,
            End_Date,
            Capacity/1000,
            Energy/1000,
            Starting_Volt,
            End_Voltage,
            Starting_current/1000,
            End_Current/1000,
            Maximum_voltage,
            Minimum_voltage,
            DCIR,
        ]
        complete_list.append(temp_list)
        
    col_list = ['Cycle Index',
                'Step Number',
                'Step Type',
                'Step Time',
                'Oneset Date',
                'End Date',
                'Capacity(Ah)',
                'Energy(Wh)',
                'Oneset Volt.(V)',
                'End Volt.(V)',
                'Starting current(A)',
                'Termination current(A)',
                'Max Volt.(V)',
                'Min Volt(V)',
                'DCIR(mΩ)']

    df=pd.DataFrame(complete_list,columns=col_list)

    dict_dtypes={
        'Step_Number':np.int64,
        'Step_Type':object,
        'Capacity(Ah)':np.float64,
        'Energy(Wh)':np.float64,
        'Starting_Voltage(V)':np.float64,
        'End_Voltage(V)':np.float64,
        'DCIR':np.float64,
        'Starting_current(A)':np.float64,
        'End_Current(A)':np.float64,
        'Maximum_voltage(V)':np.float64,
        'Minimum_voltage(V)':np.float64
    }

    # df=df.astype(dict_dtypes)

    return df

# Function to find distinct-recipes

def recipe(df):

    try:
        if type(df)!=type(pd.DataFrame()):
            df=nda_version_8_0.nda_in_df_out(df)
    except:
        raise ValueError('Arguement pushed into the function is neither a DataFrame nor a path to an nda file')
    
    dict_voltage={}
    dict_recipe={}
    dict_rest={}
    dict_current={}
    dict_step={}
    dict_cycle={}
    recipe_cycle=[]
    dict_stepid={}
    dict_cutoff_curr={}
    dict_cutoff_vol={}
    chg_temp = ''
    dchg_temp = ''
    step_col=list(df['step_name'].unique()[1:])
    for col in step_col:
        if re.search('_chg',col,re.IGNORECASE):
            chg_temp=col
        if re.search('_dchg',col,re.IGNORECASE):
            dchg_temp=col       

    for i in range(min(df['cycle']),max(df['cycle'])):
        #For non-validated files
        #if(i==7):
            #continue
        # df2=df.groupby('cycle').get_group(i)
        df2=df[df['cycle']==i]
        recipe_cycle.append(i)
        dict_recipe[i]=[]
        dict_rest[i]=[]
        dict_voltage[i]=[]
        dict_current[i]=[]
        dict_step[i]=[]
        dict_cycle[i]=[]
        dict_stepid[i]=[]
        dict_cutoff_curr[i]=[]
        dict_cutoff_vol[i]=[]
        for j in range(min(df2['step_ID']),max(df2['step_ID'])+1):
            df3=df2[df2['step_ID']==j]
            dict_recipe[i].append(df3['step_name'].iloc[0])

            dict_step[i].append(j)

            dict_cycle[i].append(i)

            dict_stepid[i].append(j)

            if(df3['step_name'].iloc[0]=='Rest'):
                dict_rest[i].append(str(timedelta(seconds=list(df3['time_in_step'])[-1])))
            else:
                dict_rest[i].append(np.nan)

            if(df3['step_name'].iloc[0]==chg_temp):
                dict_voltage[i].append(round(df3['voltage_V'].max(),2))
                dict_cutoff_vol[i].append(np.nan)
            elif(df3['step_name'].iloc[0]==dchg_temp):
                dict_voltage[i].append(round(df3['voltage_V'].min(),2))
                dict_cutoff_vol[i].append(round(list(df3['voltage_V'])[-1],2))
            else:
                dict_voltage[i].append(np.nan)
                dict_cutoff_vol[i].append(np.nan)


            if(df3['step_name'].iloc[0]==chg_temp):
                dict_current[i].append(round(df3['current_mA'].iloc[0]/1000,2))
                dict_cutoff_curr[i].append(round(list(df3['current_mA'])[-1]/1000,2))
            elif(df3['step_name'].iloc[0]==dchg_temp):
                dict_current[i].append(round(df3['current_mA'].iloc[0]/1000,2))
                dict_cutoff_curr[i].append(np.nan)
            else:
                dict_current[i].append(np.nan)
                dict_cutoff_curr[i].append(np.nan) 

    recipe=[]
    rest=[]
    voltage=[]
    current=[]
    cycle=[]
    stepid=[]
    cutoff_curr=[]
    cutoff_vol=[]

    for i in recipe_cycle:
        for j in range(len(dict_recipe[i])):
            recipe.append(dict_recipe[i][j])
            rest.append(dict_rest[i][j])
            voltage.append(dict_voltage[i][j])
            current.append(dict_current[i][j])
            cycle.append(dict_cycle[i][j])
            stepid.append(dict_stepid[i][j])  
            cutoff_curr.append(dict_cutoff_curr[i][j])
            cutoff_vol.append(dict_cutoff_vol[i][j])                     

    df_recipe=pd.DataFrame(zip(np.array(stepid).reshape(-1,1),
                           np.array(cycle).reshape(-1,1),
                           np.array(recipe).reshape(-1,1),
                           np.array(voltage).reshape(-1,1),
                           np.array(current).reshape(-1,1),
                           np.array(rest).reshape(-1,1),
                           np.array(cutoff_curr).reshape(-1,1),
                           np.array(cutoff_vol).reshape(-1,1)),columns=['Step_Id','Cycle','Step_Name','Voltage','Current','Rest','Cutoff_current','Cutoff_voltage'])                      
    
    for col in df_recipe.columns:
        df_recipe[col]=df_recipe[col].apply(lambda x : x[0]) 
        
    df_recipe.replace(np.nan,r' ',regex=True,inplace=True)
    df_recipe.replace(r'nan',r' ',regex=True,inplace=True)

    recipe_unmatch=[1]
    for i in range(len(recipe_cycle)-1):
        
        df_temp1=df_recipe[df_recipe['Cycle']==recipe_cycle[i]]
        
        df_temp2=df_recipe[df_recipe['Cycle']==recipe_cycle[i+1]]
        df_temp1=df_temp1.drop(['Cycle','Step_Id'],axis=1)
        df_temp2=df_temp2.drop(['Cycle','Step_Id'],axis=1)
        if(df_temp1.reset_index(drop=True).equals(df_temp2.reset_index(drop=True))):
            continue
        else:
            recipe_unmatch.append(recipe_cycle[i]) 
    recipe_unmatch.append(df_recipe['Cycle'].max())           
    recipe_unmatch=sorted(list(set(recipe_unmatch)))

    dict={}
    for i in range(len(recipe_unmatch)-1):
        dict[recipe_unmatch[i]]=[]
        #df1=df_recipe.groupby('Cycle').get_group(recipe_unmatch[i]).reset_index(drop=True)
        df1=df_recipe[df_recipe['Cycle']==recipe_unmatch[i]].reset_index(drop=True)
        df1=df1.drop(['Cycle','Step_Id'],axis=1)
        for j in range(i+1,len(recipe_unmatch)):
            #df2=df_recipe.groupby('Cycle').get_group(recipe_unmatch[j]).reset_index(drop=True)
            df2=df_recipe[df_recipe['Cycle']==recipe_unmatch[j]].reset_index(drop=True)
            df2=df2.drop(['Cycle','Step_Id'],axis=1)
            if(df1.equals(df2)):
                dict[recipe_unmatch[i]].append(recipe_unmatch[j]) 

    for i in dict:
        for j in dict[i]:
            if j in recipe_unmatch:
                recipe_unmatch.remove(j)
           
    arr=[]
    dict_={}
    k=0
    print(recipe_unmatch)
    for i in range(len(recipe_unmatch)):
        k+=1   
        # df_temp=df_recipe.groupby('Cycle').get_group(recipe_unmatch[i])
        df_temp=df_recipe[df_recipe['Cycle']==recipe_unmatch[i]]
        
        df_temp=df_temp.drop(['Cycle','Step_Id'],axis=1).reset_index(drop=True)
        arr.append(df_temp)
        print('Recipe-{k}'.format(k=k))
        dict_['Recipe-{k}'.format(k=k)]=df_temp
        displayhook(df_temp)
       
